/* -------------------------------------------------------------- */

/* create.rexx - prototype Web site development script 

   Paul Overaa 29th Oct 1996 

   Warning: This script shows one way of automatically generating a  
   menu driven Web site. Do not use it however unless you are familiar 
   with the script's general method of working. A tutorial on the 
   creation and management of a Web site using these types of scripts 
   has been provided in Future Publishing's Amiga Format magazine!


        SITE CREATION/UPDATE PROCEDURE
        ------------------------------
        
        1: Create all necessary text and title page files using 
           text editor. 
           
        2: Modify this script so that it includes a suitable menu
           tree definition.           
                
        3: Run the create.rexx utility using...
        
                rx create

 Some other details which may be of interest can be found in the 
 site information sections of...
            
        http://ourworld.compuserve.com/homepages/pauloveraa
   

*/

/* -------------------------------------------------------------- */

g.=''

g.LF='0A'x

g.MSG1='INSERT TITLE FOR PAGE'

g.MSG2='INSERT LINK TEXT TO PAGE'

g.MSG3='SITE UNDER DEVELOPMENT - PAGE CONTENT NOT YET AVAILABLE!'

g.MSG4='Return to page from previous level'

g.MSG5='Return directly to main index page'

g.sitedir='GeneratedSite/'
g.titlesdir='PageTitles/'
g.textdir='PageText/'

/* Tree definition follows... */

LEVELS=4

n.=-1

n.1.1=10 /* count of pages in 1st set at next level */

n.2.1=2 /* count of pages in 1st set at next level */
n.2.2=2 /* count of pages in 2nd set at next level */
n.2.3=3 /* ditto... */
n.2.4=2
n.2.5=2
n.2.6=0
n.2.7=0
n.2.8=0
n.2.9=0
n.2.10=0

n.3.1=2
n.3.2=2
n.3.3=2
n.3.4=2
n.3.5=0
n.3.6=2
n.3.7=1
n.3.8=1
n.3.9=1
n.3.10=5
n.3.11=1

/* n.4.? 16 pages at this level */

recent_page=6; index_page=10;

/* -------------------------------------------------------------- */

do level = 1 to LEVELS
        
     page_start=0
    
     say CountNodes(level) 'pages being created at level' level
         
     do node=1 to CountNodes(level)

      entries=n.level.node

      call GeneratePage(level,node,page_start,entries)
     
      page_start=page_start+entries
      
     end

end

exit

/* -------------------------------------------------------------- */

GeneratePage: Procedure expose n. g.

parse arg level,node,page_start,entries

page_name=MakeName(level,node,'htm')

page_title=MakeName(level,node,'tle')

page_text=MakeName(level,node,'txt')

parent=FindParent(level,node)

if Open(dest,g.sitedir||page_name,'w') then

  do

   if parent>0 then

      do 
      
       parent_name=MakeName(level-1,parent,'htm')
       
      end
 
   call Writeln(dest,"")
   call Writeln(dest,'<html>'||g.LF)
   call Writeln(dest,'<head>')
   call Writeln(dest,'<title> Level' level 'Page' node '</title>')
   call Writeln(dest,'</head>'||g.LF)
   call Writeln(dest,'<body>')

   if Exists(g.titlesdir||page_title) then 
   
      do 
        
        call Writech(dest,'<h2>')
        
        call InsertField(dest,g.titlesdir||page_title)
        
        call Writeln(dest,'</h2>')
      
      end      
      
      else call Writeln(dest,'<h1>' g.MSG1 page_name '</h1>')

   call Writeln(dest,'<ul>')

   target_level=level+1;

   do i=1 to entries

      target_page=page_start+i

      link_page_name=MakeName(target_level,target_page,'htm')

      link_page_title=MakeName(target_level,target_page,'tle')
   
      call Writech(dest,'<li><a href="'||link_page_name||'">')
  
      if Exists(g.titlesdir||link_page_title) then 
   
        do 

          call InsertField(dest,g.titlesdir||link_page_title)
      
          call Writech(dest,'</a>')         
   
        end      
      
      else call Writeln(dest,g.MSG2 link_page_name'</a>')

    end

   call Writeln(dest,'</ul>')

   if Exists(g.textdir||page_text) then call InsertFile(dest,g.textdir||page_text)
   
      else call Writeln(dest,g.LF||'<p>' g.MSG3'</p>'||g.LF)


   /* now do bottom of page stuff... */
   
   if parent>0 then 

      do

       call Writeln(dest,'<ul>')
   
       call Writeln(dest,'<li><a href="'||parent_name||'">'g.MSG4||'</a>')

       call Writeln(dest,'<li><a href="L01P01.htm">'||g.MSG5||'</a>')

       call Writeln(dest,'</ul>')

      end
 
   if Exists(g.textdir||page_text) then 
   
        do 

          call Writeln(dest,'<hr>')

          call Writech(dest,'<i><centre>Page last updated ')

          call Writeln(dest,'</centre></i>')
          
          call Writeln(dest,'<hr>')
        
         end

   call Writeln(dest,'</body>'||g.LF)

   call Writeln(dest,'</html>')

   CLose(dest)
   
   end

return

/* -------------------------------------------------------------- */

CountNodes: Procedure expose n.

parse arg level

level=level-1; count=0; i=1

do while n.level.i~=-1

  count=count+n.level.i
    
  i=i+1
  
end 

if count=0 then count=1
  
return count

/* -------------------------------------------------------------- */

FindParent: Procedure expose n.

parse arg level,node

if level=1 then i=0
  
  else do 
 
         level=level-1; count=0; i=0
 
         do until count>=node
         
           i=i+1
       
           count=count+n.level.i
         
         end
         
       end

return i

/* -------------------------------------------------------------- */

InsertFile: Procedure

parse arg dest, source_name

if Open(source,source_name,'r') then

  do
    
    do while ~EOF(source)

       line=Readln(source)        
        
       Writeln(dest,line)
                            
    end
     
    Close(source)
  
  end

return

/* -------------------------------------------------------------- */


InsertField: Procedure

parse arg dest, source_name

if Open(source,source_name,'r') then

  do
    
    do while ~EOF(source)

       line=Readln(source)        
        
       Writech(dest,line)
                            
    end
     
    Close(source)
  
  end

return

/* -------------------------------------------------------------- */


MakeName: Procedure

parse arg level,page,ext

if Length(level)=1 then level='0'||level

if Length(page)=1 then page='0'||page

return 'L'||level||'P'||page||'.'||ext


/* -------------------------------------------------------------- */
